# Copyright (c) 2012, Willow Garage, Inc.
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Willow Garage, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Author William Woodall/wjwwood@gmail.com

from collections import defaultdict

class Resolution(dict):
    """A default dictionary for use in the :class:`DependencyGraph`."""
    def __init__(self):
        super(Resolution, self).__init__()
        self['installer_key'] = None
        self['install_keys'] = []
        self['dependencies'] = []
        self['is_root'] = True

class DependencyGraph(defaultdict):
    """
    Provides a mechanism for generating a list of resolutions which preserves the dependency order.

    The :class:`DependencyGraph` inherits from a *defaultdict*, so it can be used as such to load the dependency graph data into it.
    Example:: 

        # Dependency graph:: A-B-C
        dg = DependencyGraph()
        dg['A']['installer_key'] = 'a_installer'
        dg['A']['install_keys'] = ['a']
        dg['A']['dependencies'] = ['B']
        dg['B']['installer_key'] = 'b_installer'
        dg['B']['install_keys'] = ['b']
        dg['B']['dependencies'] = ['C']
        dg['C']['installer_key'] = 'c_installer'
        dg['C']['install_keys'] = ['c']
        dg['C']['dependencies'] = []
        result = dg.get_ordered_uninstalled()

    """
    def __init__(self):
        defaultdict.__init__(self, Resolution)
    
    def detect_cycles(self, rosdep_key, traveled_keys):
        """
        Recursive function to detect cycles in the dependency graph.

        :param rosdep_key: This is the rosdep key to use as the root in the cycle exploration.
        :param traveled_keys: A list of rosdep_keys that have been traversed thus far.

        :raises: :exc:`AssertionError` if the rosdep_key is in the traveled keys, indicating a cycle has occurred.
        """
        assert rosdep_key not in traveled_keys, "A cycle in the dependency graph occurred with key `%s`."%rosdep_key
        traveled_keys.append(rosdep_key)
        for dependency in self[rosdep_key]['dependencies']:
            self.detect_cycles(dependency, traveled_keys)

    def validate(self):
        """
        Performs validations on the dependency graph, like cycle detection and invalid rosdep key detection. 

        :raises: :exc:`AssertionError` if a cycle is detected.
        :raises: :exc:`KeyError` if an invalid rosdep_key is found in the dependency graph.
        """
        for rosdep_key in self:
            # Ensure all dependencies have definitions
            # i.e.: Ensure we aren't pointing to invalid rosdep keys
            for dependency in self[rosdep_key]['dependencies']:
                if dependency not in self:
                    raise KeyError("Invalid Graph Structure: rosdep key `%s` does not exist in the dictionary of resolutions."%dependency)
                self[dependency]['is_root'] = False
        # Check each entry for cyclical dependencies
        for rosdep_key in self:
            self.detect_cycles(rosdep_key, [])

    def get_ordered_dependency_list(self):
        """
        Generates an ordered list of dependencies using the dependency graph.

        :returns: *[(installer_key, [install_keys])]*, ``[(str, [str])]``.  *installer_key* is the key
         that denotes which installed the accompanying *install_keys* are for.  *installer_key* are something 
         like ``apt`` or ``homebrew``.  *install_keys* are something like ``boost`` or ``ros-fuerte-ros_comm``.

        :raises: :exc:`AssertionError` if a cycle is detected.
        :raises: :exc:`KeyError` if an invalid rosdep_key is found in the dependency graph.
        """
        # Validate the graph
        self.validate()
        # Generate the dependency list
        dep_list = []
        for rosdep_key in self:
            if self[rosdep_key]['is_root']:
                dep_list.extend(self.__get_ordered_uninstalled(rosdep_key))
        # Make the list unique and remove empty entries
        result = []
        for item in dep_list:
            if item not in result and item[1] != []:
                result.append(item)
        # Squash the results by installer_key
        squashed_result = []
        previous_installer_key = None
        for installer_key, resolved in result:
            if previous_installer_key != installer_key:
                squashed_result.append((installer_key, []))
                previous_installer_key = installer_key
            squashed_result[-1][1].extend(resolved)
        return squashed_result

    def __get_ordered_uninstalled(self, key):
        uninstalled = []
        for dependency in self[key]['dependencies']:
            uninstalled.extend(self.__get_ordered_uninstalled(dependency))
        uninstalled.append((self[key]['installer_key'], self[key]['install_keys']))
        return uninstalled



# Copyright (c) 2011, Willow Garage, Inc.
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Willow Garage, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Author Ken Conley/kwc@willowgarage.com

import os
import traceback
from mock import Mock, patch

def get_test_dir():
    return os.path.abspath(os.path.join(os.path.dirname(__file__), 'osx'))

def is_port_installed_tripwire():
    # don't know the correct answer, but make sure this does not throw
    from rosdep2.platforms.osx import is_port_installed
    assert is_port_installed() in [True, False]

def is_brew_installed_tripwire():
    # don't know the correct answer, but make sure this does not throw
    from rosdep2.platforms.osx import is_brew_installed
    assert is_brew_installed() in [True, False]
    
def test_brew_detect():
    from rosdep2.platforms.osx import brew_detect
    
    m = Mock()
    m.return_value = ''
    val = brew_detect([], exec_fn=m)
    assert val == [], val

    m = Mock()
    m.return_value = ''
    val = brew_detect(['tinyxml'], exec_fn=m)
    assert val == [], val
    # make sure our test harness is based on the same implementation
    m.assert_called_with(['brew', 'list'])

    with open(os.path.join(get_test_dir(), 'brew-list-output'), 'r') as f:
        m.return_value = f.read()
    val = brew_detect(['apt', 'subversion', 'python', 'bazaar'], exec_fn=m)
    # make sure it preserves order
    assert set(val) == set(['subversion', 'bazaar'])
    assert len(val) == len(set(val))

def test_HomebrewInstaller():
    from rosdep2.platforms.osx import HomebrewInstaller

    @patch('rosdep2.platforms.osx.is_brew_installed')
    @patch.object(HomebrewInstaller, 'remove_duplicate_dependencies')
    @patch.object(HomebrewInstaller, 'get_packages_to_install')
    def test(mock_get_packages_to_install, mock_remove_duplicate_dependencies, mock_brew_installed):
        mock_brew_installed.return_value = True
        
        installer = HomebrewInstaller()
        mock_get_packages_to_install.return_value = []
        mock_remove_duplicate_dependencies.return_value = mock_get_packages_to_install.return_value
        assert [] == installer.get_install_command(['fake'])

        mock_get_packages_to_install.return_value = ['subversion', 'bazaar']
        mock_remove_duplicate_dependencies.return_value = mock_get_packages_to_install.return_value
        expected = [['brew', 'install', 'subversion'],
                    ['brew', 'install', 'bazaar']]
        # brew is always non-interactive
        for interactive in [True, False]:
            val = installer.get_install_command(['whatever'], interactive=interactive)
            
        assert val == expected, val
        expected = [['brew', 'uninstall', '--force', 'subversion'],
                    ['brew', 'install', 'subversion'],
                    ['brew', 'uninstall', '--force', 'bazaar'],
                    ['brew', 'install', 'bazaar']]
        val = installer.get_install_command(['whatever'], reinstall=True)
        assert val == expected, val
    try:
        test()
    except AssertionError:
        traceback.print_exc()
        raise
    



# Copyright (c) 2011, Willow Garage, Inc.
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Willow Garage, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Author Ken Conley/kwc@willowgarage.com

"""
Base API for loading rosdep information by package or stack name.
This API is decoupled from the ROS packaging system to enable multiple
implementations of rosdep, including ones that don't rely on the ROS
packaging system.  This is necessary, for example, to implement a
version of rosdep that works against tarballs of released stacks.
"""

import yaml

from .core import InvalidData

ROSDEP_YAML = 'rosdep.yaml'

class RosdepLoader:
    """
    Base API for loading rosdep information by package or stack name.  
    """
    
    def load_rosdep_yaml(self, yaml_contents, origin):
        """
        Utility routine for unmarshalling rosdep data encoded as YAML.

        :param origin: origin of yaml contents (for error messages)
        :raises: :exc:`yaml.YAMLError`
        """
        try:
            return yaml.load(yaml_contents)
        except yaml.YAMLError as e:
            raise InvalidData("Invalid YAML in [%s]: %s"%(origin, e), origin=origin)

    def load_view(self, view_name, rosdep_db, verbose=False):
        """
        Load view data into rosdep_db. If the view has already been
        loaded into rosdep_db, this method does nothing.

        :param view_name: name of ROS stack to load, ``str``
        :param rosdep_db: database to load stack data into, :class:`RosdepDatabase`

        :raises: :exc:`InvalidData`
        :raises: :exc:`rospkg.ResourceNotFound` if view cannot be located
        """
        raise NotImplementedError(view_name, rosdep_db, verbose) #pychecker

    def get_loadable_resources(self):
        raise NotImplementedError()

    def get_loadable_views(self):
        raise NotImplementedError()

    def get_rosdeps(self, resource_name, implicit=True):
        """
        :raises: :exc:`rospkg.ResourceNotFound` if *resource_name* cannot be found.
        """
        raise NotImplementedError(resource_name, implicit) #pychecker
    
    def get_view_key(self, resource_name):
        """
        Map *resource_name* to a view key.  In rospkg, this maps a ROS
        package name to a ROS stack name.  If *resource_name* is a ROS
        stack name, it returns the ROS stack name.

        :returns: Name of view that *resource_name* is in, ``None`` if no associated view.
        :raises: :exc:`rospkg.ResourceNotFound` if *resource_name* cannot be found.
        """
        raise NotImplementedError(resource_name) 



#!/usr/bin/env python
# Copyright (c) 2009, Willow Garage, Inc.
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Willow Garage, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Author Tully Foote/tfoote@willowgarage.com

import subprocess

from rospkg.os_detect import OS_RHEL, OS_FEDORA

from .pip import PIP_INSTALLER
from .source import SOURCE_INSTALLER
from ..installers import PackageManagerInstaller, TYPE_CODENAME
from ..shell_utils import read_stdout

# yum package manager key
YUM_INSTALLER='yum'

def register_installers(context):
    context.set_installer(YUM_INSTALLER, YumInstaller())

def register_platforms(context):
    register_fedora(context)
    register_rhel(context)
    
def register_fedora(context):
    context.add_os_installer_key(OS_FEDORA, PIP_INSTALLER)
    context.add_os_installer_key(OS_FEDORA, YUM_INSTALLER)
    context.add_os_installer_key(OS_FEDORA, SOURCE_INSTALLER)
    context.set_default_os_installer_key(OS_FEDORA, YUM_INSTALLER)
    context.set_os_version_type(OS_FEDORA, TYPE_CODENAME)

def register_rhel(context):
    context.add_os_installer_key(OS_RHEL, PIP_INSTALLER)
    context.add_os_installer_key(OS_RHEL, YUM_INSTALLER)
    context.add_os_installer_key(OS_RHEL, SOURCE_INSTALLER)
    context.set_default_os_installer_key(OS_RHEL, YUM_INSTALLER)

def rpm_detect(packages, exec_fn=None):
    ret_list = []
    #cmd = ['rpm', '-q', '--qf ""']  # suppress output for installed packages
    cmd = ['rpm', '-q', '--whatprovides', '--qf', '[%{PROVIDES}\n]']  # output: "pkg_name" for installed, error text for not installed packages
    cmd.extend(packages)

    if exec_fn is None:
        exec_fn = read_stdout

    std_out = exec_fn(cmd)
    out_lines = std_out.split('\n')
    for line in out_lines:
        # if there is no space, it's not an error text -> it's installed
        if line and ' ' not in line:
            ret_list.append(line)
    return ret_list

class YumInstaller(PackageManagerInstaller):
    """
    This class provides the functions for installing using yum
    it's methods partially implement the Rosdep OS api to complement 
    the roslib.OSDetect API.
    """

    def __init__(self):
        super(YumInstaller, self).__init__(rpm_detect)

    def get_install_command(self, resolved, interactive=True, reinstall=False):
        packages = self.get_packages_to_install(resolved, reinstall=reinstall)
        if not packages:
            return []
        elif not interactive:
            return [['sudo', 'yum', '-y', '--skip-broken', 'install'] + packages]
        else:
            return [['sudo', 'yum', '--skip-broken', 'install'] + packages]




# Copyright (c) 2012, Willow Garage, Inc.
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Willow Garage, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Author William Woodall/wjwwood@gmail.com

def test_DependencyGraph_Linear():
	from rosdep2.dependency_graph import DependencyGraph
	# Normal A-B-C
	dg = DependencyGraph()
	dg['A']['installer_key'] = 'a_installer'
	dg['A']['install_keys'] = ['a']
	dg['A']['dependencies'] = ['B']
	dg['B']['installer_key'] = 'b_installer'
	dg['B']['install_keys'] = ['b']
	dg['B']['dependencies'] = ['C']
	dg['C']['installer_key'] = 'c_installer'
	dg['C']['install_keys'] = ['c']
	dg['C']['dependencies'] = []
	result = dg.get_ordered_dependency_list()
	expected = [('c_installer', ['c']), ('b_installer', ['b']), ('a_installer', ['a'])]
	assert result == expected, "Results did not match expectations: %s == %s"%(str(result),str(expected))

def test_DependencyGraph_Cycle():
	from rosdep2.dependency_graph import DependencyGraph
	# Full Loop A-B-C-A-...
	dg = DependencyGraph()
	dg['A']['installer_key'] = 'a_installer'
	dg['A']['install_keys'] = ['a']
	dg['A']['dependencies'] = ['B']
	dg['B']['installer_key'] = 'b_installer'
	dg['B']['install_keys'] = ['b']
	dg['B']['dependencies'] = ['C']
	dg['C']['installer_key'] = 'c_installer'
	dg['C']['install_keys'] = ['c']
	dg['C']['dependencies'] = ['A']
	try:
		result = dg.get_ordered_dependency_list()
		assert False, "Doesn't fail, it should fail with an AssertionError because of the cycle."
	except AssertionError as e:
		if not str(e).startswith("A cycle in the dependency graph occurred with key"):
			assert False, "Throws AssertionError, but with the wrong message. Error was: %s: %s"%(type(e),str(e))
	except Exception as e:
		assert False, "Throws and Exception, but not an AssertionError. Error was: %s: %s"%(type(e),str(e))

def test_DependencyGraph_Short_Cycle():
	from rosdep2.dependency_graph import DependencyGraph
	# Short cycle A-B-C-D-B-C-D-...
	dg = DependencyGraph()
	dg['A']['installer_key'] = 'a_installer'
	dg['A']['install_keys'] = ['a']
	dg['A']['dependencies'] = ['B']
	dg['B']['installer_key'] = 'b_installer'
	dg['B']['install_keys'] = ['b']
	dg['B']['dependencies'] = ['C']
	dg['C']['installer_key'] = 'c_installer'
	dg['C']['install_keys'] = ['c']
	dg['C']['dependencies'] = ['D']
	dg['D']['installer_key'] = 'd_installer'
	dg['D']['install_keys'] = ['d']
	dg['D']['dependencies'] = ['B']
	try:
		result = dg.get_ordered_dependency_list()
		assert False, "Doesn't fail, it should fail with an AssertionError because of the cycle."
	except AssertionError as e:
		if not str(e).startswith("A cycle in the dependency graph occurred with key"):
			assert False, "Throws AssertionError, but with the wrong message. Error was: %s: %s"%(type(e),str(e))
	except Exception as e:
		assert False, "Throws and Exception, but not an AssertionError. Error was: %s: %s"%(type(e),str(e))

def test_DependencyGraph_Invalid_Key():
	from rosdep2.dependency_graph import DependencyGraph
	# Invalid graph A-B-C where C doesn't exist
	dg = DependencyGraph()
	dg['A']['installer_key'] = 'a_installer'
	dg['A']['install_keys'] = ['a']
	dg['A']['dependencies'] = ['B']
	dg['B']['installer_key'] = 'b_installer'
	dg['B']['install_keys'] = ['b']
	dg['B']['dependencies'] = ['C']
	try:
		result = dg.get_ordered_dependency_list()
		assert False, "Doesn't fail, it should fail with an KeyError because of the invalid rosdep key."
	except KeyError as e:
		if not str(e).endswith("does not exist in the dictionary of resolutions.'"):
			assert False, "Throws KeyError, but with the wrong message. Error was: %s: %s"%(type(e),str(e))
	except Exception as e:
		assert False, "Throws and Exception, but not an KeyError. Error was: %s: %s"%(type(e),str(e))

def test_DependencyGraph_Invalid_Key2():
	from rosdep2.dependency_graph import DependencyGraph
	# Invalid graph A-B-C where B doesn't exist
	dg = DependencyGraph()
	dg['A']['installer_key'] = 'a_installer'
	dg['A']['install_keys'] = ['a']
	dg['A']['dependencies'] = ['B']
	dg['C']['installer_key'] = 'c_installer'
	dg['C']['install_keys'] = ['c']
	dg['C']['dependencies'] = []
	try:
		result = dg.get_ordered_dependency_list()
		assert False, "Doesn't fail, it should fail with an KeyError because of the invalid rosdep key."
	except KeyError as e:
		if not str(e).endswith("does not exist in the dictionary of resolutions.'"):
			assert False, "Throws KeyError, but with the wrong message. Error was: %s: %s"%(type(e),str(e))
	except Exception as e:
		assert False, "Throws and Exception, but not an KeyError. Error was: %s: %s"%(type(e),str(e))

def test_DependencyGraph_Multi_Root():
	from rosdep2.dependency_graph import DependencyGraph
	# Multi root, shared dependency: A-B-C, D-C
	dg = DependencyGraph()
	dg['A']['installer_key'] = 'a_installer'
	dg['A']['install_keys'] = ['a']
	dg['A']['dependencies'] = ['B']
	dg['B']['installer_key'] = 'b_installer'
	dg['B']['install_keys'] = ['b']
	dg['B']['dependencies'] = ['C']
	dg['C']['installer_key'] = 'c_installer'
	dg['C']['install_keys'] = ['c']
	dg['C']['dependencies'] = []
	dg['D']['installer_key'] = 'd_installer'
	dg['D']['install_keys'] = ['d']
	dg['D']['dependencies'] = ['C']
	result = dg.get_ordered_dependency_list()
	# TODO: The expected might also have a different order, for example it might be:
	# [('c_installer', ['c']), ('d_installer', ['d']), ('b_installer', ['b']), ('a_installer', ['a'])]
	# But that wont invalidate the order from a dependency graph stand point
	expected = [
		[('c_installer', ['c']), ('b_installer', ['b']), ('a_installer', ['a']), ('d_installer', ['d'])],
		[('c_installer', ['c']), ('d_installer', ['d']), ('b_installer', ['b']), ('a_installer', ['a'])],
	]
	assert result in expected, "Results did not match expectations: %s == %s"%(str(result),str(expected))

def test_DependencyGraph_Realworld():
	from rosdep2.dependency_graph import DependencyGraph
	# Real world example
	dg = DependencyGraph()
	dg['python-matplotlib']['installer_key'] = 'pip'
	dg['python-matplotlib']['install_keys'] = ['matplotlib']
	dg['python-matplotlib']['dependencies'] = ['pkg-config']
	dg['pkg-config']['installer_key'] = 'homebrew'
	dg['pkg-config']['install_keys'] = ['pkg-config']
	dg['pkg-config']['dependencies'] = []
	result = dg.get_ordered_dependency_list()
	expected = [('homebrew', ['pkg-config']), ('pip', ['matplotlib'])]
	assert result == expected, "Results did not match expectations: %s == %s"%(str(result),str(expected))




from rosdep2.catkin_support import get_installer, get_catkin_view, ValidationFailed, resolve_for_os

from rosdep2.platforms.debian import APT_INSTALLER


def test_workflow():
    try:
        installer = get_installer(APT_INSTALLER)
        view = get_catkin_view('fuerte', 'ubuntu', 'lucid')
        resolved = resolve_for_os('cmake', view, installer, 'ubuntu', 'lucid')
        assert ['cmake'] == resolved
        resolved = resolve_for_os('python', view, installer, 'ubuntu', 'lucid')
        assert resolved == ['python-dev']
    except ValidationFailed:
        # tests fail on the server because 'rosdep init' has not been run
        pass



# Copyright (c) 2012, Willow Garage, Inc.
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Willow Garage, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
import sys
try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO

import rospkg
import rospkg.os_detect

import unittest

GITHUB_BASE_URL = 'https://github.com/ros/rosdistro/raw/master/rosdep/base.yaml'
GITHUB_PYTHON_URL = 'https://github.com/ros/rosdistro/raw/master/rosdep/python.yaml'

def get_test_dir():
    return os.path.abspath(os.path.dirname(__file__))

def get_test_tree_dir():
    return os.path.abspath(os.path.join(get_test_dir(), 'tree'))

def get_cache_dir():
    p = os.path.join(get_test_dir(), 'sources_cache')
    assert os.path.isdir(p)
    return p

from rosdep2.main import rosdep_main

from contextlib import contextmanager
@contextmanager
def fakeout():
    realstdout = sys.stdout
    realstderr = sys.stderr
    fakestdout = StringIO()
    fakestderr = StringIO()
    sys.stdout = fakestdout
    sys.stderr = fakestderr
    yield fakestdout, fakestderr
    sys.stdout = realstdout
    sys.stderr = realstderr
    
# the goal of these tests is only to test that we are wired into the
# APIs.  More exhaustive tests are at the unit level.
class TestRosdepMain(unittest.TestCase):
    def setUp(self):
        if 'ROSDEP_DEBUG' in os.environ:
            del os.environ['ROSDEP_DEBUG']
        self.old_rr = rospkg.get_ros_root()
        self.old_rpp = rospkg.get_ros_package_path()
        if 'ROS_ROOT' in os.environ:
            del os.environ['ROS_ROOT']
        os.environ['ROS_PACKAGE_PATH'] = os.path.join(get_test_tree_dir())

    def tearDown(self):
        if self.old_rr is not None:
            os.environ['ROS_ROOT'] = self.old_rr
        if self.old_rpp is not None:
            os.environ['ROS_PACKAGE_PATH'] = self.old_rpp

    def test_bad_commands(self):
        sources_cache = get_cache_dir()
        cmd_extras = ['-c', sources_cache]
        for commands in [[], ['fake', 'something'], ['check'], ['install', '-a', 'rospack_fake'],
                         ['check', 'rospack_fake', '--os', 'ubuntulucid'],
                         ]:
            try:
                rosdep_main(commands+cmd_extras)
                assert False, "system exit should have occurred"
            except SystemExit:
                pass
        
    def test_check(self):
        sources_cache = get_cache_dir()
        cmd_extras = ['-c', sources_cache]

        with fakeout() as b:
            try:
                rosdep_main(['check', 'python_dep']+cmd_extras)
            except SystemExit:
                assert False, "system exit occurred: %s\n%s"%(b[0].getvalue(), b[1].getvalue())

            stdout, stderr = b
            assert stdout.getvalue().strip() == "All system dependencies have been satisified", stdout.getvalue()
            assert not stderr.getvalue(), stderr.getvalue()
        try:
            osd = rospkg.os_detect.OsDetect()
            override = "%s:%s"%(osd.get_name(), osd.get_codename())
            with fakeout() as b:
                rosdep_main(['check', 'python_dep', '--os', override]+cmd_extras)
                stdout, stderr = b
                assert stdout.getvalue().strip() == "All system dependencies have been satisified"
                assert not stderr.getvalue(), stderr.getvalue()
        except SystemExit:
            assert False, "system exit occurred"

        # this used to abort, but now rosdep assumes validity for even empty stack args
        try:
            with fakeout() as b:
                rosdep_main(['check', 'packageless']+cmd_extras)
                stdout, stderr = b
                assert stdout.getvalue().strip() == "All system dependencies have been satisified"
                assert not stderr.getvalue(), stderr.getvalue()
        except SystemExit:
            assert False, "system exit occurred"

        try:
            rosdep_main(['check', 'nonexistent']+cmd_extras)
            assert False, "system exit should have occurred"
        except SystemExit:
            pass

    def test_install(self):
        sources_cache = get_cache_dir()
        cmd_extras = ['-c', sources_cache]

        try:
            # python must have already been installed
            with fakeout() as b:
                rosdep_main(['install', 'python_dep']+cmd_extras)
                stdout, stderr = b
                assert "All required rosdeps installed" in stdout.getvalue(), stdout.getvalue()
                assert not stderr.getvalue(), stderr.getvalue()
            with fakeout() as b:
                rosdep_main(['install', 'python_dep', '-r']+cmd_extras)
                stdout, stderr = b
                assert "All required rosdeps installed" in stdout.getvalue(), stdout.getvalue()
                assert not stderr.getvalue(), stderr.getvalue()
        except SystemExit:
            assert False, "system exit occurred: "+b[1].getvalue()
        try:
            rosdep_main(['check', 'nonexistent'])
            assert False, "system exit should have occurred"
        except SystemExit:
            pass

    def test_where_defined(self):
        try:
            sources_cache = get_cache_dir()
            expected = GITHUB_PYTHON_URL 
            for command in (['where_defined', 'testpython'], ['where_defined', 'testpython']):
                with fakeout() as b:
                    # set os to ubuntu so this test works on different platforms
                    rosdep_main(command + ['-c', sources_cache, '--os=ubuntu:lucid'])
                    stdout, stderr = b
                    output = stdout.getvalue().strip()
                    assert output == expected, output
        except SystemExit:
            assert False, "system exit occurred"
        
    def test_what_needs(self):
        try:
            sources_cache = get_cache_dir()
            cmd_extras = ['-c', sources_cache]
            expected = ['python_dep']
            with fakeout() as b:
                rosdep_main(['what-needs', 'testpython']+cmd_extras)
                stdout, stderr = b
                output = stdout.getvalue().strip()
                assert output.split('\n') == expected
            expected = ['python_dep']
            with fakeout() as b:
                rosdep_main(['what_needs', 'testpython', '--os', 'ubuntu:lucid', '--verbose']+cmd_extras)
                stdout, stderr = b
                output = stdout.getvalue().strip()
                assert output.split('\n') == expected
        except SystemExit:
            assert False, "system exit occurred"

    def test_keys(self):
        sources_cache = get_cache_dir()
        cmd_extras = ['-c', sources_cache]

        try:
            with fakeout() as b:
                rosdep_main(['keys', 'rospack_fake']+cmd_extras)
                stdout, stderr = b
                assert stdout.getvalue().strip() == "testtinyxml", stdout.getvalue()
                assert not stderr.getvalue(), stderr.getvalue()
            with fakeout() as b:
                rosdep_main(['keys', 'rospack_fake', '--os', 'ubuntu:lucid', '--verbose']+cmd_extras)
                stdout, stderr = b
                assert stdout.getvalue().strip() == "testtinyxml", stdout.getvalue()
        except SystemExit:
            assert False, "system exit occurred"
        try:
            rosdep_main(['keys', 'nonexistent']+cmd_extras)
            assert False, "system exit should have occurred"
        except SystemExit:
            pass



# Copyright (c) 2012, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Willow Garage, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
try:
    from urllib.request import urlopen
except ImportError:
    from urllib2 import urlopen


def get_test_dir():
    return os.path.abspath(os.path.join(os.path.dirname(__file__),
                                        'sources.list.d'))


def test_url_constants():
    from rosdep2.gbpdistro_support import FUERTE_GBPDISTRO_URL
    for url_name, url in [
                          ('FUERTE_GBPDISTRO_URL', FUERTE_GBPDISTRO_URL)]:
        try:
            f = urlopen(url)
            f.read()
            f.close()
        except:
            assert False, "URL [%s][%s] failed to download" % (url_name, url)


def test_get_gbprepo_as_rosdep_data():
    from rosdep2.rosdistrohelper import get_index
    from rosdep2.gbpdistro_support import get_gbprepo_as_rosdep_data
    distro = list(get_index().distributions.keys())[0]
    data = get_gbprepo_as_rosdep_data(distro)
    for k in ['ros', 'catkin', 'genmsg']:
        assert k in data, data
    assert data['ros']['ubuntu']

    try:
        get_gbprepo_as_rosdep_data('fooNonExistantDistro')
        assert False, "should have raised"
    except RuntimeError:
        pass


def test_download_gbpdistro_as_rosdep_data():
    from rosdep2.gbpdistro_support import download_gbpdistro_as_rosdep_data
    from rosdep2.gbpdistro_support import FUERTE_GBPDISTRO_URL
    from rosdep2.rep3 import REP3_TARGETS_URL
    from rosdep2 import DownloadFailure
    data = download_gbpdistro_as_rosdep_data(FUERTE_GBPDISTRO_URL)
    # don't go beyond this, this test is just making sure the download
    # plumbing is correct, not the loader.
    for k in ['ros', 'catkin', 'genmsg']:
        assert k in data, data
    assert data['ros']['ubuntu']

    # try with bad url to trigger exception handling
    try:
        # override targets URL with bad URL
        download_gbpdistro_as_rosdep_data(FUERTE_GBPDISTRO_URL,
            targets_url='http://bad.ros.org/foo.yaml')
        assert False, "should have raised"
    except DownloadFailure:
        pass
    try:
        # use targets URL, which should have a bad format
        download_gbpdistro_as_rosdep_data(REP3_TARGETS_URL)
        assert False, "should have raised"
    except DownloadFailure:
        pass


def test_gbprepo_to_rosdep_data_on_bad_inputs():
    from rosdep2.gbpdistro_support import gbprepo_to_rosdep_data
    from rosdep2 import InvalidData
    simple_gbpdistro = {'release-name': 'foorte',
                        'repositories': {},
                        'type': 'gbp'}
    targets = {'foorte': ['lucid', 'oneiric']}
    # test bad data
    try:
        gbprepo_to_rosdep_data(simple_gbpdistro, [targets])
        assert False, "should have raised"
    except InvalidData:
        pass
    try:
        gbprepo_to_rosdep_data({
            'targets': 1,
            'repositories': [],
            'type': 'gbp'}, targets)
        assert False, "should have raised"
    except InvalidData:
        pass
    try:
        gbprepo_to_rosdep_data([], targets)
        assert False, "should have raised"
    except InvalidData:
        pass
    # release-name must be in targets
    try:
        gbprepo_to_rosdep_data({
            'release-name': 'barte',
            'repositories': [],
            'type': 'gbp'}, targets)
        assert False, "should have raised"
    except InvalidData:
        pass
    # gbp-distros must be list of dicts
    try:
        gbprepo_to_rosdep_data({
            'release-name': 'foorte',
            'repositories': [1],
            'type': 'gbp'}, targets)
        assert False, "should have raised"
    except InvalidData:
        pass
    # gbp-distro target must be 'all' or a list of strings
    try:
        bad_example = {'name': 'common',
                       'target': [1],
                       'url': 'git://github.com/wg-debs/common_msgs.git'}
        gbprepo_to_rosdep_data({
            'release-name': 'foorte',
            'repositories': [bad_example],
            'type': 'gbp'}, targets)
        assert False, "should have raised"
    except InvalidData:
        pass


def test_gbprepo_to_rosdep_data_on_ok_input():
    from rosdep2.gbpdistro_support import gbprepo_to_rosdep_data
    simple_gbpdistro = {'release-name': 'foorte',
                        'repositories': {},
                        'type': 'gbp'}
    targets = {'foorte': ['lucid', 'oneiric']}
    # make sure our sample files work for the above checks before
    # proceeding to real data
    rosdep_data = gbprepo_to_rosdep_data(simple_gbpdistro, targets)
    assert rosdep_data is not None
    assert {} == rosdep_data

    gbpdistro_data = {'release-name': 'foorte',
                      'repositories': {
                          'common_msgs': dict(
                               target='all',
                               url='git://github.com/wg-debs/common_msgs.git',
                               packages={ 'foo': 'subdir/foo', 'bar': 'subdir/bar' }),
                          'gazebo': dict(
                               target=['lucid', 'natty'],
                               url='git://github.com/wg-debs/gazebo.git'),
                          'foo-bar': dict(
                               target=['precise'],
                               url='git://github.com/wg-debs/gazebo.git',
                               packages={ 'foo-bar': None }),
                          },
                      'type': 'gbp',
                      }

    rosdep_data = gbprepo_to_rosdep_data(gbpdistro_data, targets)
    for k in ['foo', 'bar', 'gazebo', 'foo-bar']:
        assert k in rosdep_data, k

    # all targets and name transform
    # These are from the 'common_msgs' repo above.
    pkgs = ['foo', 'bar']
    v = 'ros-foorte-%s'
    for pkg in pkgs:
        for p in ['lucid', 'oneiric']:
            rule = rosdep_data[pkg]['ubuntu'][p]
            assert rule['apt']['packages'] == [v % pkg], rule['apt']['packages']
        for p in ['maverick', 'natty']:
            assert p not in rosdep_data[k]['ubuntu']

    # target overrides
    pkg = 'gazebo'
    v = 'ros-foorte-gazebo'
    for p in ['lucid', 'natty']:
        rule = rosdep_data[pkg]['ubuntu'][p]
        assert rule['apt']['packages'] == [v], rule['apt']['packages']
    for p in ['oneiric', 'precise']:
        assert p not in rosdep_data[pkg]['ubuntu']

    # target overrides
    # These are from the 'foo-bar' repo above.
    v = 'ros-foorte-foo-bar'
    for pkg in ['foo-bar']:
        for p in ['precise']:
            rule = rosdep_data[pkg]['ubuntu'][p]
            assert rule['apt']['packages'] == [v], rule['apt']['packages']
        for p in ['oneiric', 'natty', 'lucid']:
            assert p not in rosdep_data[pkg]['ubuntu']


def test_get_owner_name_homebrew():
    from rosdep2.gbpdistro_support import get_owner_name
    empty_url = ''
    assert get_owner_name(empty_url) == 'ros', 'url: ' + empty_url
    https_test_url = 'https://github.com/' \
                    + 'ros/rosdistro/raw/master/releases/fuerte.yaml'
    assert get_owner_name(https_test_url) == 'ros', 'url: ' + https_test_url
    user_test_url = 'https://github.com/' \
                    + 'zklapow/rosdistro/raw/master/releases/fuerte.yaml'
    assert get_owner_name(user_test_url) == 'zklapow', 'url: ' + user_test_url
    non_github_url = 'https://ros.org/files/releases/fuerte.yaml'
    assert get_owner_name(non_github_url) == 'ros', 'url: ' + non_github_url

